;; This file directs the generation of the CLJS test namespace. The main CLJ
;; and CLJS clojure.tools.cli files are hand-crafted.
;;
;; http://birdseyesoftware.github.io/lein-dalap.docs/articles/getting_started.html#specifying_which_files_you_want_to_transform_to_cljs

{["src/test/clojure/clojure/tools/cli_test.clj"
  "src/test/clojure/cljs/tools/cli_test.cljs"]
 []}
